# Utility functions used for guest installation
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import random
import re
from subprocess import *
import tempfile
import commands

from log import log

def run(args):
    try:
        log.debug("Running: %s" % str(args))
        p = Popen(args=args, stdin=None, stdout=PIPE, stderr=STDOUT, close_fds=True,
                  env={'PATH': '/sbin:/usr/sbin:/bin:/usr/bin', 'TERM':'vt100'})
        (pid, status) = os.waitpid(p.pid, 0)
        output = p.stdout.read()
        if status:
            raise RuntimeError(output)
    except OSError, e:
        raise RuntimeError(str(e))
    return output

def close_fds(fds=xrange(0,256)):
    for fd in fds:
        try:
            os.close(fd)
        except OSError:
            pass
    if 0 in fds:
        os.open('/dev/null', os.O_RDWR)
    if 1 in fds:
        os.open('/dev/null', os.O_RDWR)
    if 2 in fds:
        os.open('/dev/null', os.O_RDWR)

def daemonize(fds=xrange(0,256), cb=None):
    pid = os.fork()
    if pid == 0:
        os.setsid()
        pid = os.fork()
        if pid == 0:
            os.chdir('/')
            os.umask(022)
            if cb:
                cb()
            close_fds(fds)
            return True
        else:
            os._exit(0)
    else:
        return False

# Word-wraps the text, and handles unicode.
# See http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/148061
rx=re.compile(u"([\u2e80-\uffff])", re.UNICODE)
def wrap(text, width=79, encoding="utf8"):
     return reduce(lambda line, word, width=width: '%s%s%s' %
                (line,
                 [' ','\n', ''][(len(line)-line.rfind('\n')-1
                       + len(word.split('\n',1)[0] ) >= width) or
                      line[-1:] == '\0' and 2],
                 word),
                rx.sub(r'\1\0 ', unicode(text,encoding)).split(' ')
            ).replace('\0', '').encode(encoding)

# Numeric-aware sorting of strings
# Inspired by http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/135435
def _generate_index(str):
    """Splits a string into alpha and numeric elements, which
    is used as an index for sorting"""
    index = []
    def _append(fragment, alist=index):
        if fragment.isdigit():
            fragment = int(fragment)
        alist.append(fragment)
    prev_isdigit = str[0].isdigit()
    current_fragment = ''
    for char in str:
        curr_isdigit = char.isdigit()
        if curr_isdigit == prev_isdigit:
            current_fragment += char
        else:
            _append(current_fragment)
            current_fragment = char
            prev_isdigit = curr_isdigit
    _append(current_fragment)    
    return tuple(index)

def sort_tuples_by_nth(alist, nth=0):
    indices = map(_generate_index, [t[nth] for t in alist])
    decorated = zip(indices, alist)
    decorated.sort()
    return [item[1] for item in decorated]

def sort_numeric_aware(alist):
    indices = map(_generate_index, alist)
    decorated = zip(indices, alist)
    decorated.sort()
    return [item for index, item in decorated]

def sort_dict_keys_by_value(d):
    sorted = sort_tuples_by_nth(d.items(), 1)
    return [ item[0] for item in sorted ]

def filtered_progress(fn, step=0.01):
    """Returns a function which takes a fraction (0.0 .. 1.0) describing
    progress.  When the progress advances beyond a specified step size,
    the function calls 'fn' with the current progress."""
    def p(f):
        if f < p.f + step:
            return
        p.f = f
        fn(p.f)
    p.f = 0.0
    fn(p.f)
    return p

def copy_to_temp(fileobj, prefix, progress=lambda f:None):
    (fd, fn) = tempfile.mkstemp(prefix=prefix)
    log.debug("Copying to temporary file '%s'" % fn)
    read = size = 0
    try:
        size = os.fstat(fileobj.fileno()).st_size
    except:
        pass
    block_size = 65536
    while 1:
        try:
            buff = fileobj.read(block_size)
            if not buff:
                break
            os.write(fd, buff)
        except:
            os.close(fd)
            os.unlink(fn)
            raise
        read += len(buff)
        if read <= size:  # Some file objects can't accurately report size
            progress(float(read) / float(size))
    os.close(fd)
    return fn

def mke2fsImage(filename, progress=lambda f:None):
    args = ['/sbin/mkfs.ext2', '-F', '-m', '0', '-q', '--', filename]
    log.debug("Creating filesystem: %s" % str(args))
    run(args)
    progress(1.0)

def mkdosfsImage(filename, progress=lambda f:None):
    args = ['/sbin/mkdosfs', '--', filename]
    log.debug("Creating filesystem: %s" % str(args))
    run(args)
    progress(1.0)

# the following three functions are from xend/uuid.py and are thus
# available under the LGPL,
# Copyright 2005 Mike Wray <mike.wray@hp.com>
# Copyright 2005 XenSource Ltd
def randomUUID():
    """Generate a random UUID."""
    return [ random.randint(0, 255) for _ in xrange(0, 16) ]

def uuidToString(u):
    return '-'.join(['%02x' * 4, '%02x' * 2, '%02x' * 2, '%02x' * 2,
                     '%02x' * 6]) % tuple(u)

def uuidFromString(s):
    s = s.replace('-', '')
    return [ int(s[i : i + 2], 16) for i in xrange(0, 32, 2) ]

def addInstallSourceRepo(source=None):
    if not source or type(source) is not str:
        return
    if source.startswith('ftp://') or source.startswith('http://') or source.startswith('nfs://') or source.startswith('smb://'):
        found = 0
        locations = commands.getoutput("/usr/bin/zypper lr -u | awk -F'|' '{ print $6 }'")
        locations = locations[(locations.rfind('URI')):].split()
        for loc in locations:
            if loc == source:
                found = 1
                break
        if not found:
            names = commands.getoutput("/usr/bin/zypper lr -u | awk -F'|' '{ print $3 }'")
            names = names[(names.rfind('Name')):].split()
            index = 1
            repo_name = "Xen-" + str(index)
            for nam in names:
                if nam == repo_name:
                    index += 1
                    repo_name = "Xen-" + str(index)
            commands.getoutput("zypper ar -d -C %s %s" % (source, repo_name))

def getDom0InstallSource():
    if os.path.exists('/var/lib/YaST2/install.inf'):
        server_ip = server_name = server_dir = inst_mode = None
        f = open('/var/lib/YaST2/install.inf')
        lines = f.readlines()
        f.close()
        # Newer install.inf files use RepoURL.  Older versions require parsing more fields
        for line in lines:
            if line.startswith('RepoURL:'):
                repo_url = line[:-1].split('?', 1)[0]
                repo_url = repo_url.split(' ')
                return repo_url[1]
            elif line.startswith('InstMode:'):
                inst_mode = line[:-1].split('?', 1)[0]
                inst_mode = inst_mode.split(' ')
                inst_mode = inst_mode[1]
                if inst_mode != 'ftp' and inst_mode != 'http' and inst_mode != 'smb' and inst_mode != 'nfs':
                    return None
            elif line.startswith('ServerIP:'):
                server_ip = line[:-1].split('?', 1)[0]
                server_ip = server_ip.split(' ')
                server_ip = server_ip[1]
            elif line.startswith('ServerName:'):
                server_name = line[:-1].split('?', 1)[0]
                server_name = server_name.split(' ')
                server_name = server_name[1]
            elif line.startswith('Serverdir:'):
                server_dir = line[:-1].split('?', 1)[0]
                server_dir = server_dir.split(' ')
                server_dir = server_dir[1]
        if inst_mode:
            repo_url = inst_mode + "://"
            if server_name:
                repo_url = repo_url + server_name + "/"
                if server_dir:
                    repo_url = repo_url + server_dir
            elif server_ip:
                repo_url = repo_url + server_ip + "/"
                if server_dir:
                    repo_url = repo_url + server_dir
            return repo_url
    return None

def getInstallSources(enabled_sources_only = False):
    dom0_inst_source = getDom0InstallSource()
    locations = commands.getoutput("/usr/bin/zypper lr -u | awk -F'|' '{ print $6 }'")
    locations = locations[(locations.rfind('URI')):].split()
    index = 0
    index_dom0 = -1
    number_of_sources = 0
    zypper_output = []
    # If we only want to list enabled sources
    if enabled_sources_only == True:
        enabled = commands.getoutput("/usr/bin/zypper lr -u | awk -F'|' '{ print $4 }'")
        enabled = enabled[(enabled.rfind('Enabled')):].split()
        for e in enabled:
            if e == "Yes":
                str = locations[index]
                if str.startswith('ftp://') or str.startswith('http://') or str.startswith('nfs://') or str.startswith('smb://'):
                    zypper_output.append(str)
                    if dom0_inst_source is not None and str == dom0_inst_source:
                        index_dom0 = number_of_sources
                    number_of_sources += 1
            index += 1
    else:
        locations.sort()
        for l in locations:
            str = locations[index]
            if str.startswith('ftp://') or str.startswith('http://') or str.startswith('nfs://') or str.startswith('smb://'):
                zypper_output.append(str)
                if dom0_inst_source is not None and str == dom0_inst_source:
                    index_dom0 = number_of_sources
                number_of_sources += 1
            index += 1
    if index_dom0 == -1 and dom0_inst_source:
        index_dom0 = 0
        zypper_output.insert(0, dom0_inst_source)
    return (index_dom0, zypper_output)
