# Communicates with xend
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import time
import libvirt
from xen.xend import XendOptions

try:
    # Xen 3.1+
    from xen.util.xmlrpcclient import ServerProxy
except:
    # Xen 3.0.4 and older
    from xen.util.xmlrpclib2 import ServerProxy

# Some Xen limits:
max_nics = 3

connection = None

def parse_string(uri):
    if not uri.startswith("xen"):
        raise ParamError(err=ParamError.E_INVALID, details=uri)
    return Xend(uri)

class Xend:
    """Thin abstraction to talk to Xend.

    DEPRECATED.  Migrate to libvirt instead.
    """
    XML_RPC_SOCKET = "/var/run/xend/xmlrpc.sock"

    def __init__(self, uri):
        if uri is None:
            self.uri = 'httpu://%s' % Xend.XML_RPC_SOCKET
            self.server = ServerProxy(self.uri)
            self.use_libvirt = False
        else:
            self.uri = uri
            try:
                self.virConn = libvirt.open(self.uri)
            except:
                self.virConn = libvirt.openReadOnly(self.uri)
            self.virNode = None
            self.virDom = None
            self.use_libvirt = True

    def info(self):
        if self.use_libvirt is False:
            return dict(self.server.xend.node.info())
        else:
            self.virNode = self.virConn.getInfo()
            # FIXME Need to calculate memory here
            return dict([('nr_cpus', self.virNode[2]), ('total_memory', self.virNode[1]), ('max_hvm_memory', 4096), ('max_para_memory', 4096), ('host', self.virConn.getHostname())])

    def getConn(self):
        if self.use_libvirt is True:
            return self.virConn
        else:
            return None

    def getURI(self):
        return self.uri

    def lookup(self, domid):
        """Lookup the domain by name, UUID, or hypervisor ID"""
        if self.use_libvirt is True:
            if type(domid) == int:
                virDom = self.virConn.lookupByID(domid)
            else:
                try:
                    virDom = self.virConn.lookupByName(domid)
                except:
                    try:
                        virDom = self.virConn.lookupByUUID(domid)
                    except:
                        return None
            return virDom
        else:
            return self.server.xend.domain(domid, False)
            
    def delete(self, domid):
        """Destroys a running instance; erases from xend"""
        try:
            if self.use_libvirt is False:
                self.server.xend.domain.delete(domid)
            else:
                if self.virDom is None:
                    self.virDom = self.lookup(domid)
                domInfo = self.virDom.info()
                self.virDom.destroy()
        except:
            pass

    def destroy(self, domid):
        """Destroys a running instance; does not erase from xend"""
        try:
            if self.use_libvirt is False:
                self.server.xend.domain.destroy(domid)
            else:
                if self.virDom is None:
                    self.virDom = self.lookup(domid)
                domInfo = self.virDom.info()
                self.virDom.shutdown()
        except:
            pass


    def domain(self, domid):
        try:
            if self.use_libvirt is False:
                return self.server.xend.domain(domid, False)
            else:
                try:
                    self.virDom = self.lookup(domid)
                    return self.virDom
                except:
                    return None
        except:
            return None

    def domain_running(self, domid):
        domain = self.domain(domid)
        if not domain:
            return False
        for d in domain:
            if type(d) == list and len(d) > 1 and d[0] == 'state':
                # Any state means there is an instance...
                return len(d[1]) > 1
        return False

    def wait_domain(self, domid):
        #  This is ugly (polling instead of event driven) but XenAPI doesn't
        #  yet grok unix socket.
        #
        #  If domain and 'state' are both found, then assume VM is trying
        #  to come up, so keep waiting.  Otherwise, fail.
        for _ in xrange(10):
            domain = self.domain(domid)
            if not domain:
                return False
            if self.use_libvirt is False:
                wait = False
                for d in domain:
                    if type(d) == list and len(d) > 1 and d[0] == 'state':
                        if len(d[1]) > 1 and not 'p' in d[1]:
                            return True
                        time.sleep(1.0)
                        wait = True
                        break
                if not wait:
                    break
            else:
                domInfo = domain.info()
                if domInfo[0] == 0 or domInfo[0] == 1:
                    return True
        return False

    def get_dom0_min_memory(self):
        if self.use_libvirt is False:
            return XendOptions.instance().get_dom0_min_mem()
        else:
            # FIXME
            return 0

